@@ -1,7 +1,7 @@
 /*
  * GfarmFS-FUSE for Gfarm version 2
  *
- * $Id: gfarm2fs.c 10209 2016-09-02 03:54:44Z n-soda $
+ * $Id$
  */
 
 #include "config.h"
@@ -19,6 +19,8 @@
 #include <errno.h>
 #include <stddef.h>
 #include <limits.h>
+#include <pthread.h>
+#include <assert.h>
 #ifdef HAVE_SYS_XATTR_H
 #include <sys/xattr.h>
 #endif
@@ -158,6 +160,46 @@ const static char *gfarm2fs_subdir;
 static size_t gfarm2fs_subdir_len;
 #define IS_SUBDIR(p)	(memcmp(p, gfarm2fs_subdir, gfarm2fs_subdir_len) == 0)
 
+static void
+open_file_lock_init(struct gfarm2fs_file *fp)
+{
+	int rv;
+	rv = pthread_rwlock_init(&fp->lock, NULL);
+	assert(rv == 0);
+}
+
+static void
+open_file_lock_destroy(struct gfarm2fs_file *fp)
+{
+	int rv;
+	rv = pthread_rwlock_destroy(&fp->lock);
+	assert(rv == 0);
+}
+
+static void
+open_file_rdlock(struct gfarm2fs_file *fp)
+{
+	int rv;
+	rv = pthread_rwlock_rdlock(&fp->lock);
+	assert(rv == 0);
+}
+
+static void
+open_file_wrlock(struct gfarm2fs_file *fp)
+{
+	int rv;
+	rv = pthread_rwlock_wrlock(&fp->lock);
+	assert(rv == 0);
+}
+
+static void
+open_file_unlock(struct gfarm2fs_file *fp)
+{
+	int rv;
+	rv = pthread_rwlock_unlock(&fp->lock);
+	assert(rv == 0);
+}
+
 static void
 gfarm2fs_record_mount_point(const char *mpoint, const char *subdir)
 {
@@ -441,10 +483,14 @@ gfarm2fs_fstat(
 
 	/* assert(st_outp); */
 
+	open_file_rdlock(fp);
+
 	/* get atime, mtime and size from gfsd */
 	e = gfs_pio_stat(fp->gf, st_outp); /* include gfs_fstat() */
-	if (e != GFARM_ERR_NO_ERROR)
+	if (e != GFARM_ERR_NO_ERROR) {
+		open_file_unlock(fp);
 		return (e);
+	}
 
 	if (fp->time_updated) { /* use atime and mtime from gfmd */
 		if (st_inp == NULL) {
@@ -454,6 +500,7 @@ gfarm2fs_fstat(
 			e = gfs_fstat(fp->gf, &st_gfmd); /* from gfmd */
 			if (e != GFARM_ERR_NO_ERROR) {
 				gfs_stat_free(st_outp);
+				open_file_unlock(fp);
 				return (e);
 			}
 			st_outp->st_atimespec = st_gfmd.st_atimespec;
@@ -464,6 +511,7 @@ gfarm2fs_fstat(
 			st_outp->st_mtimespec = st_inp->st_mtimespec;
 		}
 	}
+	open_file_unlock(fp);
 	return (GFARM_ERR_NO_ERROR);
 }
 
@@ -507,6 +555,7 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 		free_gfarmized_path(&gfarmized);
 		return (-gfarm_error_to_errno(e));
 	}
+	gfarm2fs_open_file_table_lock();
 	if ((fp = gfarm2fs_open_file_lookup(st.st_ino)) != NULL) {
 		struct gfs_stat st2;
 
@@ -516,11 +565,13 @@ gfarm2fs_getattr(const char *path, struct stat *stbuf)
 			gfarm2fs_check_error(GFARM_MSG_2000046, OP_GETATTR,
 				"gfs_pio_stat", gfarmized.path, e);
 			free_gfarmized_path(&gfarmized);
+			gfarm2fs_open_file_table_unlock();
 			return (-gfarm_error_to_errno(e));
 		}
 		gfs_stat_free(&st);
 		st = st2;
 	}
+	gfarm2fs_open_file_table_unlock();
 	copy_gfs_stat(gfarmized.path, stbuf, &st);
 	gfs_stat_free(&st);
 	free_gfarmized_path(&gfarmized);
@@ -1103,12 +1154,14 @@ gfarm2fs_ftruncate(const char *path, off_t size,
 	struct gfarm2fs_file *fp = get_filep(fi);
 
 	(void) path;
+	open_file_wrlock(fp);
 	e = gfs_pio_truncate(fp->gf, size);
 	if (e != GFARM_ERR_NO_ERROR) {
 		gfarm2fs_check_error(GFARM_MSG_2000024, OP_FTRUNCATE,
 		    "gfs_pio_ftruncate", path, e);
 	} else
 		fp->time_updated = 0;
+	open_file_unlock(fp);
 	return (-gfarm_error_to_errno(e));
 }
 
@@ -1134,13 +1187,17 @@ gfarm2fs_utimens(const char *path, const struct timespec ts[2])
 		free_gfarmized_path(&gfarmized);
 		return (-gfarm_error_to_errno(e));
 	}
+	gfarm2fs_open_file_table_lock();
 	if ((fp = gfarm2fs_open_file_lookup(st.st_ino)) != NULL) {
+		open_file_wrlock(fp);
 		fp->gt[0].tv_sec = ts[0].tv_sec;
 		fp->gt[0].tv_nsec = ts[0].tv_nsec;
 		fp->gt[1].tv_sec = ts[1].tv_sec;
 		fp->gt[1].tv_nsec = ts[1].tv_nsec;
 		fp->time_updated = 1;
+		open_file_unlock(fp);
 	}
+	gfarm2fs_open_file_table_unlock();
 	gfs_stat_free(&st);
 	gt[0].tv_sec = ts[0].tv_sec;
 	gt[0].tv_nsec = ts[0].tv_nsec;
@@ -1192,7 +1249,7 @@ gfs_hook_open_flags_gfarmize(int open_flags)
 	if ((open_flags & O_EXCL) != 0)
 		gfs_flags |= GFARM_FILE_EXCLUSIVE;
 #endif
-#if 0 /* not yet on Gfarm v2 */
+#ifdef GFARM_FILE_UNBUFFERED
 	/* open(2) and creat(2) should be unbuffered */
 	gfs_flags |= GFARM_FILE_UNBUFFERED;
 #endif
@@ -1217,6 +1274,7 @@ gfarm2fs_file_init(
 		fp->gf = gf;
 		fp->time_updated = 0;
 		fp->inum = st.st_ino;
+		open_file_lock_init(fp);
 		*fpp = fp;
 		gfs_stat_free(&st);
 		return (GFARM_ERR_NO_ERROR);
@@ -1307,24 +1365,19 @@ gfarm2fs_read(const char *path, char *buf, size_t size, off_t offset,
 	struct fuse_file_info *fi)
 {
 	gfarm_error_t e;
-	gfarm_off_t off;
 	int rv;
 	struct gfarm2fs_file *fp = get_filep(fi);
 
 	(void) path;
-	e = gfs_pio_seek(fp->gf, offset, GFARM_SEEK_SET, &off);
-	if (e != GFARM_ERR_NO_ERROR) {
-		gfarm2fs_check_error(GFARM_MSG_2000028, OP_READ,
-					"gfs_pio_seek", path, e);
-	} else {
-		e = gfs_pio_read(fp->gf, buf, size, &rv);
-		gfarm2fs_check_error(GFARM_MSG_2000029, OP_READ,
-					"gfs_pio_read", path, e);
-	}
+	open_file_wrlock(fp);
+	e = gfs_pio_pread(fp->gf, buf, size, offset, &rv);
+	gfarm2fs_check_error(GFARM_MSG_2000029, OP_READ,
+				"gfs_pio_read", path, e);
 	if (e != GFARM_ERR_NO_ERROR)
 		rv = -gfarm_error_to_errno(e);
 	else
 		fp->time_updated = 0;
+	open_file_unlock(fp);
 	return (rv);
 }
 
@@ -1333,24 +1386,19 @@ gfarm2fs_write(const char *path, const char *buf, size_t size,
 	off_t offset, struct fuse_file_info *fi)
 {
 	gfarm_error_t e;
-	gfarm_off_t off;
 	int rv;
 	struct gfarm2fs_file *fp = get_filep(fi);
 
 	(void) path;
-	e = gfs_pio_seek(fp->gf, offset, GFARM_SEEK_SET, &off);
-	if (e != GFARM_ERR_NO_ERROR) {
-		gfarm2fs_check_error(GFARM_MSG_2000030, OP_WRITE,
-					"gfs_pio_seek", path, e);
-	} else {
-		e = gfs_pio_write(fp->gf, buf, size, &rv);
-		gfarm2fs_check_error(GFARM_MSG_2000031, OP_WRITE,
-					"gfs_pio_write", path, e);
-	}
+	open_file_wrlock(fp);
+	e = gfs_pio_pwrite(fp->gf, buf, size, offset, &rv);
+	gfarm2fs_check_error(GFARM_MSG_2000031, OP_WRITE,
+				"gfs_pio_write", path, e);
 	if (e != GFARM_ERR_NO_ERROR)
 		rv = -gfarm_error_to_errno(e);
 	else
 		fp->time_updated = 0;
+	open_file_unlock(fp);
 	return (rv);
 }
 
@@ -1389,6 +1437,7 @@ gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 
 	(void) path;
 	gfarm2fs_open_file_remove(fp);
+	open_file_wrlock(fp);
 	e = gfs_pio_close(fp->gf);
 	gfarm2fs_check_error(GFARM_MSG_2000033, OP_RELEASE,
 				"gfs_pio_close", path, e);
@@ -1410,6 +1459,8 @@ gfarm2fs_release(const char *path, struct fuse_file_info *fi)
 			free_gfarmized_path(&gfarmized);
 		}
 	}
+	open_file_unlock(fp);
+	open_file_lock_destroy(fp);
 	free(fp);
 	return (-gfarm_error_to_errno(e));
 }
@@ -1435,16 +1486,19 @@ gfarm2fs_fsync(const char *path, int isdatasync, struct fuse_file_info *fi)
 static int
 gfarm2fs_flush(const char *path, struct fuse_file_info *fi)
 {
+	int rv = 0;
 	gfarm_error_t e;
 	struct gfarm2fs_file *fp = get_filep(fi);
 
+	open_file_rdlock(fp);
 	if (IS_WRITABLE(fp->flags)) {
 		e = gfs_pio_flush(fp->gf);
 		gfarm2fs_check_error(GFARM_MSG_2000123, OP_FLUSH,
 		    "gfs_pio_flush", path, e);
-		return (-gfarm_error_to_errno(e));
-	} else
-		return (0);
+		rv = -gfarm_error_to_errno(e);
+	}
+	open_file_unlock(fp);
+	return (rv);
 }
 
 #if defined(HAVE_SYS_XATTR_H) && defined(ENABLE_XATTR)
@@ -1517,7 +1571,11 @@ gfarm2fs_getxattr(const char *path, const char *name, char *value, size_t size)
 		 *   ...
 		 */
 		free_gfarmized_path(&gfarmized);
+#ifdef ENOATTR /* for macOS, etc */
+		return (-ENOATTR);
+#else
 		return (-ENODATA);
+#endif
 	}
 	if (e != GFARM_ERR_NO_ERROR) {
 		gfarm2fs_check_error(GFARM_MSG_2000037, OP_GETXATTR,
@@ -1570,10 +1628,15 @@ gfarm2fs_removexattr(const char *path, const char *name)
 	gfarm2fs_check_error(GFARM_MSG_2000039, OP_REMOVEXATTR,
 			     "gfs_lremovexattr", gfarmized.path, e);
 	free_gfarmized_path(&gfarmized);
-	if (e == GFARM_ERR_NO_SUCH_OBJECT)
+	if (e == GFARM_ERR_NO_SUCH_OBJECT) {
+#ifdef ENOATTR /* for macOS, etc */
+		return (-ENOATTR);
+#else
 		return (-ENODATA);
-	else
+#endif
+	} else {
 		return (-gfarm_error_to_errno(e));
+	}
 }
 #endif /* HAVE_SYS_XATTR_H && ENABLE_XATTR */
 
@@ -1654,69 +1717,53 @@ uncache_path(const char *path)
 static int
 gfarm2fs_mknod_cached(const char *path, mode_t mode, dev_t rdev)
 {
-	int rv = gfarm2fs_mknod(path, mode, rdev);
-
-	if (rv == 0)
-		uncache_parent(path);
-	return (rv);
+	/* uncache always to avoid race condition */
+	uncache_parent(path);
+	return (gfarm2fs_mknod(path, mode, rdev));
 }
 
 static int
 gfarm2fs_mkdir_cached(const char *path, mode_t mode)
 {
-	int rv = gfarm2fs_mkdir(path, mode);
-
-	if (rv == 0)
-		uncache_parent(path);
-	return (rv);
+	uncache_parent(path);
+	return (gfarm2fs_mkdir(path, mode));
 }
 
 static int
 gfarm2fs_unlink_cached(const char *path)
 {
-	int rv = gfarm2fs_unlink(path);
-
-	if (rv == 0) {
-		uncache_path(path);
-		uncache_parent(path);
-	}
-	return (rv);
+	uncache_path(path);
+	uncache_parent(path);
+	return (gfarm2fs_unlink(path));
 }
 
 static int
 gfarm2fs_rmdir_cached(const char *path)
 {
-	int rv = gfarm2fs_rmdir(path);
-
-	if (rv == 0) {
-		uncache_path(path);
-		uncache_parent(path);
-	}
-	return (rv);
+	uncache_path(path);
+	uncache_parent(path);
+	return (gfarm2fs_rmdir(path));
 }
 
 static int
 gfarm2fs_symlink_cached(const char *old, const char *to)
 {
-	int rv = gfarm2fs_symlink(old, to);
-
-	if (rv == 0)
-		uncache_parent(to);
-	return (rv);
+	uncache_parent(to);
+	return (gfarm2fs_symlink(old, to));
 }
 
 static int
 gfarm2fs_rename_cached(const char *from, const char *to)
 {
-	int rv = gfarm2fs_rename(from, to);
+	int rv;
 	struct gfs_stat st;
 
+	uncache_path(from);
+	uncache_parent(from);
+	uncache_path(to);
+	uncache_parent(to);
+	rv = gfarm2fs_rename(from, to);
 	if (rv == 0) {
-		uncache_path(from);
-		uncache_parent(from);
-		uncache_path(to);
-		uncache_parent(to);
-
 		/* try to replicate the destination file just in case */
 		if (gfs_lstat_cached(to, &st) == GFARM_ERR_NO_ERROR) {
 			if (GFARM_S_ISREG(st.st_mode))
@@ -1730,40 +1777,31 @@ gfarm2fs_rename_cached(const char *from, const char *to)
 static int
 gfarm2fs_link_cached(const char *from, const char *to)
 {
-	int rv = gfarm2fs_link(from, to);
-
-	if (rv == 0)
-		uncache_parent(to);
-	return (rv);
+	uncache_parent(to);
+	return (gfarm2fs_link(from, to));
 }
 
 static int
 gfarm2fs_chmod_cached(const char *path, mode_t mode)
 {
-	int rv = gfarm2fs_chmod(path, mode);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_chmod(path, mode));
 }
 
 static int
 gfarm2fs_chown_cached(const char *path, uid_t uid, gid_t gid)
 {
-	int rv = gfarm2fs_chown(path, uid, gid);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_chown(path, uid, gid));
 }
 
 static int
 gfarm2fs_truncate_cached(const char *path, off_t size)
 {
-	int rv = gfarm2fs_truncate(path, size);
+	int rv;
 
-	if (rv == 0)
-		uncache_path(path);
+	uncache_path(path);
+	rv = gfarm2fs_truncate(path, size);
 	gfarm2fs_replicate(path, NULL);
 	return (rv);
 }
@@ -1772,63 +1810,49 @@ static int
 gfarm2fs_ftruncate_cached(const char *path, off_t size,
 			struct fuse_file_info *fi)
 {
-	int rv = gfarm2fs_ftruncate(path, size, fi);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_ftruncate(path, size, fi));
 }
 
 static int
 gfarm2fs_utimens_cached(const char *path, const struct timespec ts[2])
 {
-	int rv = gfarm2fs_utimens(path, ts);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_utimens(path, ts));
 }
 
 static int
 gfarm2fs_create_cached(const char *path, mode_t mode, struct fuse_file_info *fi)
 {
-	int rv = gfarm2fs_create(path, mode, fi);
-
-	if (rv == 0)
-		uncache_parent(path);
-	return (rv);
+	uncache_parent(path);
+	return (gfarm2fs_create(path, mode, fi));
 }
 
 static int
 gfarm2fs_open_cached(const char *path, struct fuse_file_info *fi)
 {
-	int rv = gfarm2fs_open(path, fi);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_open(path, fi));
 }
 
 static int
 gfarm2fs_write_cached(const char *path, const char *buf, size_t size,
 	off_t offset, struct fuse_file_info *fi)
 {
-	int rv = gfarm2fs_write(path, buf, size, offset, fi);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_write(path, buf, size, offset, fi));
 }
 
 static int
 gfarm2fs_release_cached(const char *path, struct fuse_file_info *fi)
 {
-	int rv = gfarm2fs_release(path, fi);
+	int rv;
 
-	if (rv == 0 && ((fi->flags & O_ACCMODE) == O_WRONLY ||
-			(fi->flags & O_ACCMODE) == O_RDWR ||
-			(fi->flags & O_TRUNC) != 0))
+	if ((fi->flags & O_ACCMODE) == O_WRONLY ||
+	    (fi->flags & O_ACCMODE) == O_RDWR ||
+	    (fi->flags & O_TRUNC) != 0)
 		uncache_path(path);
+	rv = gfarm2fs_release(path, fi);
 	gfarm2fs_replicate(path, fi);
 	return (rv);
 }
@@ -1838,21 +1862,15 @@ static int
 gfarm2fs_setxattr_cached(const char *path, const char *name, const char *value,
 	size_t size, int flags)
 {
-	int rv = gfarm2fs_setxattr(path, name, value, size, flags);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_setxattr(path, name, value, size, flags));
 }
 
 static int
 gfarm2fs_removexattr_cached(const char *path, const char *name)
 {
-	int rv = gfarm2fs_removexattr(path, name);
-
-	if (rv == 0)
-		uncache_path(path);
-	return (rv);
+	uncache_path(path);
+	return (gfarm2fs_removexattr(path, name));
 }
 
 #endif /* HAVE_SETXATTR && ENABLE_XATTR */
@@ -2134,11 +2152,10 @@ main(int argc, char *argv[])
 		fprintf(stderr, "failed to parse option\n");
 		exit(1);
 	}
-	/*
-	 * specify '-s' option to disable multithreaded operations
-	 * libgfarm is not thread-safe for now
-	 */
+#ifndef HAVE_GFS_PROFILE_LOCK
+	/* specify '-s' option to disable multithreaded operations */
 	fuse_opt_add_arg(&args, "-s");
+#endif
 #if FUSE_VERSION >= 28
 	/* -o atomic_o_trunc required to overwrite a "lost all replica" file */
 	fuse_opt_add_arg(&args, "-oatomic_o_trunc");
